#include "cmaudio.h"

/* table to scale scale from OSS mixer value to AC97 mixer register value */
static struct ac97_mixer_hw ac97_hw_cmi9739[SOUND_MIXER_NRDEVICES]= {
	[SOUND_MIXER_VOLUME]  = { AC97_MASTER_VOL_STEREO,	32 },
	[SOUND_MIXER_PCM]     = { AC97_PCMOUT_VOL,		32 },
	[SOUND_MIXER_LINE]    = { AC97_LINEIN_VOL,		32 },
	[SOUND_MIXER_MIC]     = { AC97_MIC_VOL,			32 },
	[SOUND_MIXER_CD]      = { AC97_CD_VOL,			32 },
	[SOUND_MIXER_LINE1]   = { AC97_AUX_VOL,			32 },
};

/* this table has default mixer values for all OSS mixers. */
static struct mixer_defaults_cmi9739 {
	int mixer;
	unsigned int value;
} mixer_defaults_cmi9739[SOUND_MIXER_NRDEVICES] = {
	/* all values 0 -> 100 in bytes */
	{ SOUND_MIXER_VOLUME,	0x4949 },
	{ SOUND_MIXER_PCM,	0x4949 },
	{ SOUND_MIXER_LINE,	0x4949 },
	{ SOUND_MIXER_MIC,	0x4949 },
	{ SOUND_MIXER_CD,	0x4949 },
	{ SOUND_MIXER_LINE1,	0x4949 },
	{ -1, 0 }
};

static int RecMask_io(struct cmedia_card *card, int rw, int mask)
{
    unsigned int val, item;
    struct ac97_mixer_hw *mh;

    if (rw) {
	// check spdif in first
	val = card->ReadRegisterWord(card, AC97REG_SPDIF_FUNCTION);
	if (val & 0x10)
		return 1 << SOUND_MIXER_DIGITAL1;

	// read it from the card
	val = card->ReadRegisterWord(card, AC97_RECORD_SELECT);
	return 1 << ac97_rm2oss[val & 0x07];
    }

    /* else, write the first set in the mask as the output */
    /* clear out current set value first (AC97 supports only 1 input!) */

    // check spdif in first
    val = card->ReadRegisterWord(card, AC97REG_SPDIF_FUNCTION);
    if (val & 0x10)
	    val = (1 << SOUND_MIXER_DIGITAL1);
    else {
	val = card->ReadRegisterWord(card, AC97_RECORD_SELECT) & 0x07;
	val = 1 << ac97_rm2oss[val];
    }

    if (mask != val)
	    mask &= ~val;

    if (mask == 1 << SOUND_MIXER_DIGITAL1) {
	MaskRegOr(card, AC97REG_SPDIF_FUNCTION, 0x10);
	card->WriteRegisterWord(card, AC97_RECORD_SELECT, 0x0707);
	card->WriteRegisterWord(card, AC97_RECORD_GAIN, 0x8000);
	return 0;
    }
    else
	MaskRegAnd(card, AC97REG_SPDIF_FUNCTION, ~0x10);

    val = ffs(mask);
    val = ac97_oss_rm[val-1];
    val |= val << 8;  /* set both channels */

    card->WriteRegisterWord(card, AC97_RECORD_SELECT, val);

    item = ac97_rm2oss[val & 0x07];
    mh = &ac97_hw_cmi9739[item];
    val = card->ReadRegisterWord(card, mh->offset);
    switch (item) {
    case SOUND_MIXER_MIC:
	val = ((val & 0x001f) << 8) | (val & 0x801f);
	break;
    default:
	val &= 0x9f1f;
	break;
    }
    val = (val & 0x8000) | (~((val & 0x1f1f) >> 1) & 0x0f0f);
    card->WriteRegisterWord(card, AC97_RECORD_GAIN, val);

    return 0;
};

/* write the OSS encoded volume to the given OSS encoded mixer, again caller's job to
   make sure all is well in arg land, call with spinlock held */
static void write_mixer_9739(struct cmedia_card *card, int oss_channel, unsigned int left, unsigned int right)
{
    u16 val = 0, item;
    struct ac97_mixer_hw *mh = NULL;

    if (oss_channel == SOUND_MIXER_DIGITAL1) {
	val = card->ReadRegisterWord(card, AC97REG_SPDIF_FUNCTION);
	if (val & 0x10)
	    card->WriteRegisterWord(card, AC97_RECORD_GAIN, AC97_MUTE);

	if (left == 0 && right == 0) {
	    card->bMonitorSPDIF = 0;
	    MaskRegAnd(card, AC97REG_SPDIF_FUNCTION, ~0x08);
	} else {
	    card->bMonitorSPDIF = 1;
	    if (card->dma_state & DAC_RUNNING)
		MaskRegAnd(card, AC97REG_SPDIF_FUNCTION, ~0x08);
	    else
		MaskRegOr(card, AC97REG_SPDIF_FUNCTION, 0x08);
	}
	return;
    } else if (oss_channel == SOUND_MIXER_DIGITAL2) {
	if (left == 0 && right == 0)
	    // Disable S/PDIF out
	    MaskRegAnd(card, AC97_EXTENDED_STATUS, ~0x0004);
	else
	    // Enable S/PDIF out
	    MaskRegOr(card, AC97_EXTENDED_STATUS, 0x0004);
	return;
    }

    mh = &ac97_hw_cmi9739[oss_channel];

    if (oss_channel == SOUND_MIXER_MIC) {
	val = card->ReadRegisterWord(card, mh->offset) & ~0x801f;
	left = ((100 - left) * mh->scale) / 100;
	if (left >= mh->scale)
		left = mh->scale - 1;
	val |= left;
	/*  the low bit is optional in the tone sliders and masking
	    it lets us avoid the 0xf 'bypass'.. */
    } else {
	if (left == 0 && right == 0)
		val = AC97_MUTE;
    	else {
		right = ((100 - right) * mh->scale) / 100;
		left = ((100 - left) * mh->scale) / 100;
		if (right >= mh->scale)
			right = mh->scale - 1;
		if (left >= mh->scale)
			left = mh->scale - 1;
		val = (left << 8) | right;
	}
    }

    card->WriteRegisterWord(card, mh->offset, val);

    // check spdif in first
    val = card->ReadRegisterWord(card, AC97REG_SPDIF_FUNCTION);
    if (val & 0x10) {
	card->WriteRegisterWord(card, AC97_RECORD_GAIN, AC97_MUTE);
	return;
    }

    // Setting Software Recording Gain
    val = card->ReadRegisterWord(card, AC97_RECORD_SELECT);
    item = ac97_rm2oss[val & 0x07];
    mh = &ac97_hw_cmi9739[item];
    val = card->ReadRegisterWord(card, mh->offset);
    switch (item) {
    case SOUND_MIXER_MIC:
	if (val & 0x8000)
		val = ((val & 0x001f) << 8) | (val & 0x001f);
	break;
    default:
	val &= 0x9f1f;
	break;
    }
    val = (val & 0x8000) | (~((val & 0x1f1f) >> 1) & 0x0f0f);
    card->WriteRegisterWord(card, AC97_RECORD_GAIN, val);
}

static int Null(struct cmedia_card *card, int enable)
{
	return -EINVAL;
}

/* a thin wrapper for write_mixer */
static void Set_Mixer(struct cmedia_card *card, unsigned int oss_mixer, unsigned int val)
{
    unsigned int left, right;

    /* cleanse input a little */
    right = ((val >> 8)  & 0xff) ;
    left = (val  & 0xff) ;

    if (right > 100)
	    right = 100;
    if (left > 100)
	    left = 100;

    card->mixer_state[oss_mixer] = (right << 8) | left;
    write_mixer_9739(card, oss_mixer, left, right);
}

static int LineinAsRear(struct cmedia_card *card, int val)
{
	u8 Reg;
	u16 AndMask, OrMask;

	switch (val) {
	case 0:
	case 1:
		OrMask = 0;
		AndMask = ~0;
		card->LineInAsRear = val;
		Reg = AC97REG_MULTI_CHANNEL_CTRL;
		if (val)
			OrMask |= 0x0400;
		else
			AndMask &= ~0x0400;
		MaskReg(card, Reg, AndMask, OrMask);
		break;
	default:
		val = card->LineInAsRear;
		break;
	}
	return val;
}

static int MicAsCBass(struct cmedia_card *card, int val)
{
	u16 AndMask, OrMask;

	switch (val) {
	case 0:
	case 1:
		OrMask = 0;
		AndMask = ~0;
		card->MicAsCenterBass = val;

#if FORCE_MICSELECT2_MODE
		MaskRegOr(card, AC97_GENERAL_PURPOSE, AC97_GP_MS);
#else
		if (val)
      			MaskRegOr(card, AC97_GENERAL_PURPOSE, AC97_GP_MS);
		else
       			MaskRegAnd(card, AC97_GENERAL_PURPOSE, ~AC97_GP_MS);
#endif
		AndMask = val ? ~0x2000 : ~0x1000;
		OrMask = val ? 0x1000 : 0x2000;
		MaskReg(card, AC97REG_MULTI_CHANNEL_CTRL, AndMask, OrMask);
		break;
	default:
		val = card->MicAsCenterBass;
		break;
	}
	return val;
}

static int Xear(struct cmedia_card *card, int enable)
{
	switch (enable) {
	case 0:
	case 1:
		card->XearMode = enable;
		break;
	default:
		enable = card->XearMode;
		break;
	}
	return enable;
}

static int MicSelect2(struct cmedia_card *card, int val)
{
	switch (val) {
	case 0:
	case 1:
		if ((card->output_channels >= 6) && card->MicAsCenterBass && (val == 0))
			val = 1;
		card->MicSelect2 = val;
		if (val)
			MaskRegOr(card, AC97_GENERAL_PURPOSE, AC97_GP_MS);
		else
			MaskRegAnd(card, AC97_GENERAL_PURPOSE, ~AC97_GP_MS);
		break;
	default:
		val = card->MicSelect2;
		break;
	}
	return val;
}

static int Mix2Surround(struct cmedia_card *card, int val)
{
	switch (val) {
	case 0:
	case 1:
		if (val)
			MaskRegOr(card, AC97REG_MULTI_CHANNEL_CTRL, 0x0200);
		else
			MaskRegAnd(card, AC97REG_MULTI_CHANNEL_CTRL, ~0x0200);
		card->Mix2Surround = val;
	    	break;
	default:
		val = card->Mix2Surround;
		break;
	}
	return val;
}

static int MicBoostSel(struct cmedia_card *card, int val)
{
	switch (val) {
	case 0:
	case 1:
		if (val)
			MaskRegOr(card , AC97REG_MULTI_CHANNEL_CTRL, 1);
		else
			MaskRegAnd(card , AC97REG_MULTI_CHANNEL_CTRL, ~1);
		card->MicBoost = val;
		break;
	default:
		val = card->MicBoost;
		break;
	}
	return val;
}

static int SetMonitorSPDIF(struct cmedia_card *card, int enable)
{
	if (!card->bMonitorSPDIF)
		return enable;

	switch (enable) {
	case 0:
		MaskRegOr(card, AC97REG_SPDIF_FUNCTION, 0x08);
		card->MonitorSPDIF = enable;
		break;
	case 1:
		MaskRegAnd(card, AC97REG_SPDIF_FUNCTION, ~0x08);
		card->MonitorSPDIF = enable;
		break;
	default:
		enable = card->MonitorSPDIF;
		break;
	}
	return enable;
}

void InitialMixerCMI9739(struct cmedia_card *card)
{
    unsigned int val;
    int i;

	card->RecMask_io = RecMask_io;
	card->Set_Mixer = Set_Mixer;
	card->SetSpeakers = Null;
	card->LineAsRear = LineinAsRear;
	card->MicAsCBass = MicAsCBass;
	card->Xear = Xear;
	card->MicSel2 = MicSelect2;
	card->Mix2Surr = Mix2Surround;
	card->MicBoostSel = MicBoostSel;
	card->FMicSel = Null;
	card->SetMonitorSPDIF = SetMonitorSPDIF;

    card->supported_mixers = SOUND_MASK_VOLUME|SOUND_MASK_PCM|SOUND_MASK_LINE|SOUND_MASK_MIC|
			     SOUND_MASK_CD|SOUND_MASK_LINE1|SOUND_MASK_DIGITAL1;
    card->stereo_mixers = card->supported_mixers & ~SOUND_MASK_MIC;
    card->record_sources = card->supported_mixers & ~SOUND_MASK_PCM;

    /* initialize mixer channel volumes */
    card->WriteRegisterWord(card, AC97_RESET, 0);
    for (i = 0; i < SOUND_MIXER_NRDEVICES; i++) {
	struct mixer_defaults_cmi9739 *md = &mixer_defaults_cmi9739[i];

	if (md->mixer == -1)
		break;
	if (!card_supported_mixer(card, md->mixer))
		continue;
	Set_Mixer(card, md->mixer, md->value);
    }
    Set_Mixer(card, SOUND_MIXER_DIGITAL1, 0x6464);
    Set_Mixer(card, SOUND_MIXER_DIGITAL2, 0x6464);

    // Open Microphone
    card->WriteRegisterWord(card, AC97REG_MULTI_CHANNEL_CTRL, 0x3000);

    // Set Recording Gain
    card->WriteRegisterWord(card, AC97_RECORD_GAIN, 0x0000);

    // Set Surround Mixer Volume
    card->WriteRegisterWord(card, AC97_SURROUND_MASTER, 0x0000);

    // Set Center/Base Mixer Volume
    card->WriteRegisterWord(card, AC97_CENTER_LFE_MASTER, 0x0000);

    // Set Recording Select
    val = card->RecMask_io(card, 1, 0);
    card->RecMask_io(card, 0, val);

    // Enable S/PDIF out
    val = card->ReadRegisterWord(card, AC97_EXTENDED_STATUS);
    card->WriteRegisterWord(card, AC97_EXTENDED_STATUS, 0x05c4);
    val = card->ReadRegisterWord(card, AC97REG_SPDIF_FUNCTION);
    card->WriteRegisterWord(card, AC97REG_SPDIF_FUNCTION, 0x0009);
    card->bMonitorSPDIF=1;

    // Enable Mix2Surround
    MaskRegOr(card, AC97REG_MULTI_CHANNEL_CTRL, 0x0200);
    card->Mix2Surround = 1;
}
